from __future__ import annotations

import argparse
import logging
import sys

from biliass import Danmaku2ASS
from biliass.__version__ import VERSION as biliass_version


def main():
    logging.basicConfig(format="%(levelname)s: %(message)s")
    if len(sys.argv) == 1:
        sys.argv.append("--help")
    parser = argparse.ArgumentParser(description="bilibili ASS Danmaku converter", prog="biliass")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {biliass_version}")
    parser.add_argument("-o", "--output", metavar="OUTPUT", help="Output file")
    parser.add_argument(
        "-s",
        "--size",
        metavar="WIDTHxHEIGHT",
        required=True,
        help="Stage size in pixels",
    )
    parser.add_argument(
        "-fn",
        "--font",
        metavar="FONT",
        help="Specify font face [default: sans-serif]",
        default="sans-serif",
    )
    parser.add_argument(
        "-fs",
        "--fontsize",
        metavar="SIZE",
        help="Default font size [default: 25]",
        type=float,
        default=25.0,
    )
    parser.add_argument("-a", "--alpha", metavar="ALPHA", help="Text opacity", type=float, default=1.0)
    parser.add_argument(
        "-dm",
        "--duration-marquee",
        metavar="SECONDS",
        help="Duration of scrolling comment display [default: 5]",
        type=float,
        default=5.0,
    )
    parser.add_argument(
        "-ds",
        "--duration-still",
        metavar="SECONDS",
        help="Duration of still comment display [default: 5]",
        type=float,
        default=5.0,
    )
    parser.add_argument("-fl", "--filter", help="Regular expression to filter comments")
    parser.add_argument(
        "-flf",
        "--filter-file",
        help="Regular expressions from file (one line one regex) to filter comments",
    )
    parser.add_argument(
        "-p",
        "--protect",
        metavar="HEIGHT",
        help="Reserve blank on the bottom of the stage",
        type=int,
        default=0,
    )
    parser.add_argument(
        "-r",
        "--reduce",
        action="store_true",
        help="Reduce the amount of comments if stage is full",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["xml", "protobuf"],
        default="xml",
        help="Input danmaku format (xml or protobuf)",
    )
    parser.add_argument("file", metavar="FILE", nargs="+", help="Comment file to be processed")
    args = parser.parse_args()
    try:
        width, height = str(args.size).split("x", 1)
        width = int(width)
        height = int(height)
    except ValueError:
        raise ValueError(f"Invalid stage size: {args.size!r}") from None

    inputs: list[str | bytes] = []
    for file in args.file:
        try:
            with open(file, "r" if args.format == "xml" else "rb") as f:  # noqa: PTH123
                inputs.append(f.read())
        except UnicodeDecodeError:
            logging.error(f"Failed to decode file {file}, if it is a protobuf file, please use `-f protobuf`")
            sys.exit(1)

    if args.output:
        fout = open(args.output, "w", encoding="utf-8-sig", errors="replace", newline="\r\n")  # noqa: PTH123
    else:
        fout = sys.stdout
    output = Danmaku2ASS(
        inputs,
        width,
        height,
        args.format,
        args.protect,
        args.font,
        args.fontsize,
        args.alpha,
        args.duration_marquee,
        args.duration_still,
        args.filter,
        args.reduce,
    )
    fout.write(output)
    fout.close()


if __name__ == "__main__":
    main()



from __future__ import annotations

import shutil
from pathlib import Path
from typing import TYPE_CHECKING

from yutto.utils.fetcher import DEFAULT_HEADERS as DEFAULT_HEADERS

if TYPE_CHECKING:
    import pytest

TEST_DIR = Path("./__test_files__")


def pytest_sessionstart(session: pytest.Session):
    TEST_DIR.mkdir(exist_ok=True)


def pytest_sessionfinish(session: pytest.Session, exitstatus: int):
    if TEST_DIR.exists():
        shutil.rmtree(TEST_DIR)



from __future__ import annotations

import pytest

from yutto.processor.selector import (
    parse_episodes_selection,
    validate_episodes_selection,
)


@pytest.mark.processor
def test_regex():
    # 单个
    assert validate_episodes_selection("1")
    assert validate_episodes_selection("99")
    assert validate_episodes_selection("-1")
    assert validate_episodes_selection("-99")
    assert validate_episodes_selection("$")
    assert not validate_episodes_selection("")
    assert not validate_episodes_selection(" ")
    assert not validate_episodes_selection("x")
    assert not validate_episodes_selection("- 1")
    assert not validate_episodes_selection("1$")

    # 组合
    assert validate_episodes_selection("1,2")
    assert validate_episodes_selection("1,-2,3,-4")
    assert not validate_episodes_selection("1, 2")
    assert not validate_episodes_selection("1,")

    # 范围
    assert validate_episodes_selection("1~3")
    assert validate_episodes_selection("1~-1")
    assert validate_episodes_selection("-2~-1")
    assert not validate_episodes_selection("1~2~3")

    # 范围 + 组合
    assert validate_episodes_selection("1~2,9~$")
    assert validate_episodes_selection("0~10,12~14,-2~$")

    # 起止省略语法糖
    assert validate_episodes_selection("~2,9~")
    assert validate_episodes_selection("9~,~2")
    assert validate_episodes_selection("~")


@pytest.mark.processor
def test_single():
    assert parse_episodes_selection("1", 24) == [1]
    assert parse_episodes_selection("11", 24) == [11]
    assert parse_episodes_selection("-1", 24) == [24]
    assert parse_episodes_selection("-10", 24) == [15]
    assert parse_episodes_selection("$", 24) == [24]
    assert parse_episodes_selection("25", 24) == []


@pytest.mark.processor
def test_compose():
    assert parse_episodes_selection("1,2,4", 24) == [1, 2, 4]
    assert parse_episodes_selection("11,14,15", 24) == [11, 14, 15]
    assert parse_episodes_selection("11,14,25", 24) == [11, 14]
    assert parse_episodes_selection("11,-1,$", 24) == [11, 24]
    assert parse_episodes_selection("$,-10", 24) == [15, 24]


@pytest.mark.processor
def test_range():
    assert parse_episodes_selection("1~4", 24) == [1, 2, 3, 4]
    assert parse_episodes_selection("1~100", 6) == [1, 2, 3, 4, 5, 6]
    assert parse_episodes_selection("4~10", 6) == [4, 5, 6]
    assert parse_episodes_selection("2~-2", 6) == [2, 3, 4, 5]
    assert parse_episodes_selection("2~$", 6) == [2, 3, 4, 5, 6]


@pytest.mark.processor
def test_range_and_compose():
    assert parse_episodes_selection("1~4,6~8", 24) == [1, 2, 3, 4, 6, 7, 8]
    assert parse_episodes_selection("1~4,2~6", 24) == [1, 2, 3, 4, 5, 6]
    assert parse_episodes_selection("1~4,5~6", 24) == [1, 2, 3, 4, 5, 6]
    assert parse_episodes_selection("1~4,5~6,8", 24) == [1, 2, 3, 4, 5, 6, 8]
    assert parse_episodes_selection("3,5~7,12,17", 24) == [3, 5, 6, 7, 12, 17]
    assert parse_episodes_selection("1~3,10,12~14,16,-4~$", 24) == [1, 2, 3, 10, 12, 13, 14, 16, 21, 22, 23, 24]


@pytest.mark.processor
def test_sugar():
    assert parse_episodes_selection("~4,20~", 24) == parse_episodes_selection("1~4,20~24", 24)
    assert parse_episodes_selection("~4,20~$", 24) == parse_episodes_selection("1~4,20~24", 24)
    assert parse_episodes_selection("~", 24) == parse_episodes_selection("1~24", 24)



from __future__ import annotations

import pytest

from yutto._typing import AvId, CId
from yutto.api.danmaku import get_danmaku, get_protobuf_danmaku_segment, get_xml_danmaku
from yutto.utils.fetcher import create_client
from yutto.utils.funcutils import as_sync


@pytest.mark.api
@as_sync
async def test_xml_danmaku():
    cid = CId("144541892")
    async with create_client() as client:
        danmaku = await get_xml_danmaku(client, cid=cid)
        assert len(danmaku) > 0


@pytest.mark.api
@as_sync
async def test_protobuf_danmaku():
    cid = CId("144541892")
    async with create_client() as client:
        danmaku = await get_protobuf_danmaku_segment(client, cid=cid, segment_id=1)
        assert len(danmaku) > 0


@pytest.mark.api
@as_sync
async def test_danmaku():
    cid = CId("144541892")
    avid = AvId("BV1q7411v7Vd")
    async with create_client() as client:
        danmaku = await get_danmaku(client, cid=cid, avid=avid, save_type="ass")
        assert len(danmaku["data"]) > 0
        assert danmaku["source_type"] == "xml"
        assert danmaku["save_type"] == "ass"



from __future__ import annotations

import pytest

from yutto.api.user_info import get_user_info
from yutto.utils.fetcher import create_client
from yutto.utils.funcutils import as_sync


@pytest.mark.api
@as_sync
async def test_get_user_info():
    async with create_client() as client:
        user_info = await get_user_info(client)
        assert not user_info["vip_status"]
        assert not user_info["is_login"]



from __future__ import annotations

import subprocess
import sys

import pytest

from yutto.__version__ import VERSION as yutto_version

from .conftest import TEST_DIR

PYTHON = sys.executable


@pytest.mark.e2e
def test_version_e2e():
    p = subprocess.run([PYTHON, "-m", "yutto", "-v"], capture_output=True, check=True)
    res = p.stdout.decode()
    assert res.strip().endswith(yutto_version)


@pytest.mark.e2e
@pytest.mark.ci_skip
def test_bangumi_e2e():
    short_bangumi = "https://www.bilibili.com/bangumi/play/ep100367"
    subprocess.run(
        [PYTHON, "-m", "yutto", short_bangumi, f"-d={TEST_DIR}", "-q=16", "-w"],
        capture_output=True,
        check=True,
    )


@pytest.mark.e2e
def test_ugc_video_e2e():
    short_ugc_video = "https://www.bilibili.com/video/BV1AZ4y147Yg"
    subprocess.run(
        [PYTHON, "-m", "yutto", short_ugc_video, f"-d={TEST_DIR}", "-q=16", "-w"],
        capture_output=True,
        check=True,
    )



from __future__ import annotations

# import pytest
from yutto.utils.ffmpeg import FFmpegCommandBuilder


def test_video_input_only():
    command_builder = FFmpegCommandBuilder()
    command_builder.add_video_input("input.m4s")
    command_builder.add_output("output.mp4")
    excepted_command = ["-i", "input.m4s", "--", "output.mp4"]
    assert command_builder.build() == excepted_command


def test_audio_input_only():
    command_builder = FFmpegCommandBuilder()
    command_builder.add_audio_input("input.aac")
    command_builder.add_output("output.mp4")
    excepted_command = ["-i", "input.aac", "--", "output.mp4"]
    assert command_builder.build() == excepted_command


def test_merge_video_audio_with_auto_stream_selection():
    command_builder = FFmpegCommandBuilder()
    command_builder.add_video_input("input.m4s")
    command_builder.add_audio_input("input.aac")
    command_builder.add_output("output.mp4")
    excepted_command = ["-i", "input.m4s", "-i", "input.aac", "--", "output.mp4"]
    assert command_builder.build() == excepted_command


def test_merge_video_audio_with_manual_stream_selection_select_all():
    command_builder = FFmpegCommandBuilder()
    video_input = command_builder.add_video_input("input.m4s")
    audio_input = command_builder.add_audio_input("input.aac")
    output = command_builder.add_output("output.mp4")
    output.use(video_input)
    output.use(audio_input)
    excepted_command = ["-i", "input.m4s", "-i", "input.aac", "-map", "0", "-map", "1", "--", "output.mp4"]
    assert command_builder.build() == excepted_command


def test_merge_video_audio_with_manual_stream_selection_select_video_only():
    command_builder = FFmpegCommandBuilder()
    video_input = command_builder.add_video_input("input.m4s")
    command_builder.add_audio_input("input.aac")
    output = command_builder.add_output("output.mp4")
    output.use(video_input)
    excepted_command = ["-i", "input.m4s", "-i", "input.aac", "-map", "0", "--", "output.mp4"]
    assert command_builder.build() == excepted_command


def test_merge_video_audio_with_cover():
    command_builder = FFmpegCommandBuilder()
    video_input = command_builder.add_video_input("input.m4s")
    audio_input = command_builder.add_audio_input("input.aac")
    cover_input = command_builder.add_video_input("cover.jpg")
    output = command_builder.add_output("output.mp4")
    output.use(video_input)
    output.use(audio_input)
    output.use(cover_input)
    output.set_cover(cover_input)
    excepted_command = [
        "-i",
        "input.m4s",
        "-i",
        "input.aac",
        "-i",
        "cover.jpg",
        "-map",
        "0",
        "-map",
        "1",
        "-map",
        "2",
        "-c:v:1",
        "copy",
        "-disposition:v:1",
        "attached_pic",
        "--",
        "output.mp4",
    ]
    assert command_builder.build() == excepted_command


def test_merge_video_audio_with_cover_reorder():
    command_builder = FFmpegCommandBuilder()
    cover_input = command_builder.add_video_input("cover.jpg")
    video_input = command_builder.add_video_input("input.m4s")
    audio_input = command_builder.add_audio_input("input.aac")
    output = command_builder.add_output("output.mp4")
    output.use(cover_input)
    output.use(audio_input)
    output.use(video_input)
    output.set_cover(cover_input)
    excepted_command = [
        "-i",
        "cover.jpg",
        "-i",
        "input.m4s",
        "-i",
        "input.aac",
        "-map",
        "0",
        "-map",
        "2",
        "-map",
        "1",
        "-c:v:0",
        "copy",
        "-disposition:v:0",
        "attached_pic",
        "--",
        "output.mp4",
    ]
    assert command_builder.build() == excepted_command


def test_merge_video_audio_with_codec():
    command_builder = FFmpegCommandBuilder()
    command_builder.add_video_input("input.m4s")
    command_builder.add_audio_input("input.aac")
    output = command_builder.add_output("output.mp4")
    output.set_vcodec("hevc")
    output.set_acodec("copy")
    excepted_command = [
        "-i",
        "input.m4s",
        "-i",
        "input.aac",
        "-vcodec",
        "hevc",
        "-acodec",
        "copy",
        "--",
        "output.mp4",
    ]
    assert command_builder.build() == excepted_command


def test_merge_video_audio_with_extra_options():
    command_builder = FFmpegCommandBuilder()
    command_builder.add_video_input("input.m4s")
    command_builder.add_audio_input("input.aac")
    output = command_builder.add_output("output.mp4")
    output.with_extra_options(["-strict", "unofficial"])
    command_builder.with_extra_options(["-threads", "8"])
    excepted_command = [
        "-i",
        "input.m4s",
        "-i",
        "input.aac",
        "-threads",
        "8",
        "-strict",
        "unofficial",
        "--",
        "output.mp4",
    ]
    assert command_builder.build() == excepted_command



from __future__ import annotations

from typing import TYPE_CHECKING

import httpx
import pytest
from biliass import read_comments_bilibili_xml

from ..conftest import DEFAULT_HEADERS, TEST_DIR

if TYPE_CHECKING:
    from pathlib import Path


def gen_xml_v1(base_dir: Path):
    filename = "test_v1.xml"
    filepath = base_dir / filename
    cid = "18678311"
    resp = httpx.get(
        f"http://comment.bilibili.com/{cid}.xml",
        headers=DEFAULT_HEADERS,
        follow_redirects=True,
    )
    resp.encoding = "utf-8"
    with filepath.open("w", encoding="utf-8") as f:
        f.write(resp.text)


@pytest.mark.biliass
def test_xml_v1_text():
    gen_xml_v1(TEST_DIR)
    with TEST_DIR.joinpath("test_v1.xml").open("r") as f:
        list(read_comments_bilibili_xml(f.read(), 10))


@pytest.mark.biliass
def test_xml_v1_bytes():
    gen_xml_v1(TEST_DIR)
    with TEST_DIR.joinpath("test_v1.xml").open("rb") as f:
        list(read_comments_bilibili_xml(f.read(), 10))



from __future__ import annotations

from typing import Any

import pytest

from yutto.utils.funcutils.data_access import Undefined, data_has_chained_keys

TEST_DATA: list[tuple[Any, list[str], bool]] = [
    # basic
    ({"a": None}, ["a"], True),
    ({"a": Undefined()}, ["a"], False),
    ({"a": Undefined()}, ["a", "b"], False),
    ({"a": 1}, ["a"], True),
    ({"a": 1}, ["a", "b"], False),
    ({"a": 1}, ["b"], False),
    # nested
    ({"a": {"b": 1}}, ["a"], True),
    ({"a": {"b": 1}}, ["a", "b"], True),
    ({"a": {"b": 1}}, ["a", "b", "c"], False),
    ({"a": {"b": 1}}, ["a", "c", "b"], False),
    ({"a": {"b": 1}}, ["0", "1", "2"], False),
    ({"a": {"b": None}}, ["a", "b"], True),
    # not a dict
    (None, [], True),
    (1, [], True),
]


@pytest.mark.parametrize("data, keys, expected", TEST_DATA)
def test_data_has_chained_keys(data: Any, keys: list[str], expected: bool):
    assert data_has_chained_keys(data, keys) == expected



from __future__ import annotations

from biliass._core import get_danmaku_meta_size as get_danmaku_meta_size

from .biliass import (
    Danmaku2ASS as Danmaku2ASS,
    read_comments_bilibili_protobuf as read_comments_bilibili_protobuf,
    read_comments_bilibili_xml as read_comments_bilibili_xml,
)



from __future__ import annotations

from typing import TYPE_CHECKING

import httpx
import pytest
from biliass import read_comments_bilibili_protobuf

from ..conftest import DEFAULT_HEADERS, TEST_DIR

if TYPE_CHECKING:
    from pathlib import Path


def gen_protobuf(base_dir: Path):
    filename = "test.pb"
    filepath = base_dir / filename
    cid = "18678311"
    resp = httpx.get(
        f"http://api.bilibili.com/x/v2/dm/web/seg.so?type=1&oid={cid}&segment_index={1}",
        headers=DEFAULT_HEADERS,
    )
    with filepath.open("wb") as f:
        f.write(resp.content)


@pytest.mark.biliass
def test_protobuf():
    gen_protobuf(TEST_DIR)
    with TEST_DIR.joinpath("test.pb").open("rb") as f:
        list(read_comments_bilibili_protobuf(f.read(), 10))



from __future__ import annotations

import pytest

from yutto._typing import BvId, MId, SeriesId
from yutto.api.collection import get_collection_details
from yutto.utils.fetcher import create_client
from yutto.utils.funcutils import as_sync


@pytest.mark.api
@as_sync
async def test_get_collection_details():
    # 测试页面：https://space.bilibili.com/6762654/channel/collectiondetail?sid=39879&ctype=0
    series_id = SeriesId("39879")
    mid = MId("6762654")
    async with create_client() as client:
        collection_details = await get_collection_details(client, series_id=series_id, mid=mid)
        title = collection_details["title"]
        avids = [page["avid"] for page in collection_details["pages"]]
        assert title == "傻开心整活"
        assert BvId("BV1er4y1H7tQ") in avids
        assert BvId("BV1Yi4y1C7u6") in avids



from __future__ import annotations

import argparse
import asyncio
import copy
import os
import re
import sys
from typing import TYPE_CHECKING, Any, Callable, Literal

import httpx
from typing_extensions import TypeAlias

from yutto.__version__ import VERSION as yutto_version
from yutto.bilibili_typing.quality import (
    audio_quality_priority_default,
    video_quality_priority_default,
)
from yutto.exceptions import ErrorCode
from yutto.extractor import (
    BangumiBatchExtractor,
    BangumiExtractor,
    CheeseBatchExtractor,
    CheeseExtractor,
    CollectionExtractor,
    FavouritesExtractor,
    SeriesExtractor,
    UgcVideoBatchExtractor,
    UgcVideoExtractor,
    UserAllFavouritesExtractor,
    UserAllUgcVideosExtractor,
    UserWatchLaterExtractor,
)
from yutto.processor.downloader import DownloadState, start_downloader
from yutto.processor.parser import alias_parser, file_scheme_parser
from yutto.processor.path_resolver import create_unique_path_resolver
from yutto.utils.asynclib import sleep_with_status_bar_refresh
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.fetcher import Fetcher, create_client
from yutto.utils.funcutils import as_sync
from yutto.utils.time import TIME_DATE_FMT, TIME_FULL_FMT
from yutto.validator import (
    initial_validation,
    validate_basic_arguments,
    validate_batch_arguments,
    validate_user_info,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from yutto._typing import EpisodeData

DownloadResourceType: TypeAlias = Literal["video", "audio", "subtitle", "metadata", "danmaku", "cover", "chapter_info"]
DOWNLOAD_RESOURCE_TYPES: list[DownloadResourceType] = [
    "video",
    "audio",
    "subtitle",
    "metadata",
    "danmaku",
    "cover",
    "chapter_info",
]


def main():
    parser = cli()
    args = parser.parse_args()
    initial_validation(args)
    args_list = flatten_args(args, parser)
    try:
        run(args_list)
    except (SystemExit, KeyboardInterrupt, asyncio.exceptions.CancelledError):
        Logger.info("已终止下载，再次运行即可继续下载～")
        sys.exit(ErrorCode.PAUSED_DOWNLOAD.value)


def cli() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="yutto 一个可爱且任性的 B 站视频下载器", prog="yutto")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {yutto_version}")
    # 如果需要创建其他子命令可参考
    # https://stackoverflow.com/questions/29998417/create-parser-with-subcommands-in-argparse-customize-positional-arguments
    parser.add_argument("url", help="视频主页 url 或 url 列表（需使用 file scheme）")
    group_common = parser.add_argument_group("common", "通用参数")
    group_common.add_argument("-n", "--num-workers", type=int, default=8, help="同时用于下载的最大 Worker 数")
    group_common.add_argument(
        "-q",
        "--video-quality",
        default=127,
        choices=video_quality_priority_default,
        type=int,
        help="视频清晰度等级（127:8K, 126:Dolby Vision, 125:HDR, 120:4K, 116:1080P60, 112:1080P+, 100:智能修复, 80:1080P, 74:720P60, 64:720P, 32:480P, 16:360P）",
    )
    group_common.add_argument(
        "-aq",
        "--audio-quality",
        default=30251,
        choices=audio_quality_priority_default,
        type=int,
        help="音频码率等级（30251:Hi-Res, 30255:Dolby Audio, 30250:Dolby Atmos, 30280:320kbps, 30232:128kbps, 30216:64kbps）",
    )
    group_common.add_argument(
        "--vcodec",
        default="avc:copy",
        metavar="DOWNLOAD_VCODEC:SAVE_VCODEC",
        help="视频编码格式（<下载格式>:<生成格式>）",
    )
    group_common.add_argument(
        "--acodec",
        default="mp4a:copy",
        metavar="DOWNLOAD_ACODEC:SAVE_ACODEC",
        help="音频编码格式（<下载格式>:<生成格式>）",
    )
    group_common.add_argument(
        "--download-vcodec-priority",
        default="auto",
        help="视频编码格式优先级，使用 `,` 分隔，如 `hevc,avc,av1`，默认为 `auto`，即根据 vcodec 中「下载编码」自动推断",
    )
    group_common.add_argument(
        "--output-format", default="infer", choices=["infer", "mp4", "mkv", "mov"], help="输出格式（infer 为自动推断）"
    )
    group_common.add_argument(
        "--output-format-audio-only",
        default="infer",
        choices=["infer", "aac", "mp3", "flac", "mp4", "mkv", "mov"],
        help="仅包含音频流时所使用的输出格式（infer 为自动推断）",
    )
    group_common.add_argument(
        "-df", "--danmaku-format", default="ass", choices=["xml", "ass", "protobuf"], help="弹幕类型"
    )
    group_common.add_argument(
        "-bs", "--block-size", default=0.5, type=float, help="分块下载时各块大小，单位为 MiB，默认为 0.5MiB"
    )
    group_common.add_argument("-w", "--overwrite", action="store_true", help="强制覆盖已下载内容")
    group_common.add_argument(
        "-x", "--proxy", default="auto", help="设置代理（auto 为系统代理、no 为不使用代理、当然也可以设置代理值）"
    )
    group_common.add_argument("-d", "--dir", default="./", help="下载目录，默认为运行目录")
    group_common.add_argument("--tmp-dir", help="用来存放下载过程中临时文件的目录，默认为下载目录")
    group_common.add_argument("-c", "--sessdata", default="", help="Cookies 中的 SESSDATA 字段")
    group_common.add_argument("-tp", "--subpath-template", default="{auto}", help="多级目录的存储路径模板")
    group_common.add_argument(
        "-af", "--alias-file", type=argparse.FileType("r", encoding="utf-8"), help="设置 url 别名文件路径"
    )
    group_common.add_argument(
        "--metadata-format-premiered", default=TIME_DATE_FMT, help="专用于 metadata 文件中 premiered 字段的日期格式"
    )
    group_common.add_argument("--download-interval", default=0, type=int, help="设置下载间隔，单位为秒")
    group_common.add_argument("--banned-mirrors-pattern", default=None, help="禁用下载链接的镜像源，使用正则匹配")

    # 资源选择
    group_common.add_argument(
        "--video-only",
        dest="require_audio",
        action=create_select_required_action(deselect=["audio"]),
        help="仅下载视频流",
    )
    group_common.add_argument(
        "--audio-only",
        dest="require_video",
        action=create_select_required_action(deselect=["video"]),
        help="仅下载音频流",
    )  # 视频和音频是反选对方，而不是其余反选所有的
    group_common.add_argument(
        "--no-danmaku",
        dest="require_danmaku",
        action=create_select_required_action(deselect=["danmaku"]),
        help="不生成弹幕文件",
    )
    group_common.add_argument(
        "--danmaku-only",
        dest="require_danmaku",
        action=create_select_required_action(select=["danmaku"], deselect=invert_selection(["danmaku"])),
        help="仅生成弹幕文件",
    )
    group_common.add_argument(
        "--no-subtitle",
        dest="require_subtitle",
        action=create_select_required_action(deselect=["subtitle"]),
        help="不生成字幕文件",
    )
    group_common.add_argument(
        "--subtitle-only",
        dest="require_subtitle",
        action=create_select_required_action(select=["subtitle"], deselect=invert_selection(["subtitle"])),
        help="仅生成字幕文件",
    )
    group_common.add_argument(
        "--with-metadata",
        dest="require_metadata",
        action=create_select_required_action(select=["metadata"]),
        help="生成元数据文件",
    )
    group_common.add_argument(
        "--metadata-only",
        dest="require_metadata",
        action=create_select_required_action(select=["metadata"], deselect=invert_selection(["metadata"])),
        help="仅生成元数据文件",
    )
    group_common.add_argument(
        "--no-cover",
        dest="require_cover",
        action=create_select_required_action(deselect=["cover"]),
        help="不生成封面",
    )

    group_common.add_argument(
        "--no-chapter-info",
        dest="require_chapter_info",
        action=create_select_required_action(deselect=["chapter_info"]),
        help="不封装章节信息",
    )

    group_common.set_defaults(
        require_video=True,
        require_audio=True,
        require_subtitle=True,
        require_metadata=False,
        require_danmaku=True,
        require_cover=True,
        require_chapter_info=True,
    )
    group_common.add_argument("--no-color", action="store_true", help="不使用颜色")
    group_common.add_argument("--no-progress", action="store_true", help="不显示进度条")
    group_common.add_argument("--debug", action="store_true", help="启用 debug 模式")
    group_common.add_argument("--vip-strict", action="store_true", help="启用严格检查大会员生效")
    group_common.add_argument("--login-strict", action="store_true", help="启用严格检查登录状态")

    # 仅批量下载使用
    group_batch = parser.add_argument_group("batch", "批量下载参数")
    group_batch.add_argument("-b", "--batch", action="store_true", help="批量下载")
    group_batch.add_argument("-p", "--episodes", default="1~-1", help="选集")
    group_batch.add_argument(
        "-s", "--with-section", action="store_true", help="同时下载附加剧集（PV、预告以及特别篇等专区内容）"
    )
    group_batch.add_argument("--batch-filter-start-time", help="只下载该时间之后（包含临界值）发布的稿件")
    group_batch.add_argument("--batch-filter-end-time", help="只下载该时间之前（不包含临界值）发布的稿件")

    # 仅任务列表中使用
    group_batch_file = parser.add_argument_group("batch file", "批量下载文件参数")
    group_batch_file.add_argument("--no-inherit", action="store_true", help="不继承父级参数")

    return parser


@as_sync
async def run(args_list: list[argparse.Namespace]):
    unique_path = create_unique_path_resolver()
    async with create_client(
        cookies=Fetcher.cookies,
        trust_env=Fetcher.trust_env,
        proxy=Fetcher.proxy,
    ) as client:
        if len(args_list) > 1:
            Logger.info(f"列表里共检测到 {len(args_list)} 项")

        for i, args in enumerate(args_list):
            if len(args_list) > 1:
                Logger.custom(f"列表项 {args.url}", Badge(f"[{i+1}/{len(args_list)}]", fore="black", back="cyan"))

            # 验证批量参数
            if args.batch:
                validate_batch_arguments(args)

            # 初始化各种提取器
            extractors = (
                [
                    UgcVideoBatchExtractor(),  # 投稿全集
                    BangumiBatchExtractor(),  # 番剧全集
                    CheeseBatchExtractor(),  # 课程全集
                    FavouritesExtractor(),  # 用户单一收藏
                    UserAllFavouritesExtractor(),  # 用户全部收藏
                    SeriesExtractor(),  # 视频列表
                    CollectionExtractor(),  # 视频合集
                    UserAllUgcVideosExtractor(),  # 个人空间，由于个人空间的正则包含了收藏夹，所以需要放在收藏夹之后
                    UserWatchLaterExtractor(),  # 用户稍后再看
                ]
                if args.batch
                else [
                    UgcVideoExtractor(),  # 投稿单集
                    BangumiExtractor(),  # 番剧单话
                    CheeseExtractor(),  # 课程单集
                ]
            )
            url: str = args.url
            # 将 shortcut 转为完整 url
            for extractor in extractors:
                matched, url = extractor.resolve_shortcut(url)
                if matched:
                    break

            # 在开始前校验，减少对第一个视频的请求
            if not await validate_user_info({"is_login": args.login_strict, "vip_status": args.vip_strict}):
                Logger.error("启用了严格校验大会员或登录模式，请检查 SESSDATA 或大会员状态！")
                sys.exit(ErrorCode.NOT_LOGIN_ERROR.value)
            # 重定向到可识别的 url
            try:
                url = await Fetcher.get_redirected_url(client, url)
            except httpx.InvalidURL:
                Logger.error(f"无效的 url({url})～请检查一下链接是否正确～")
                sys.exit(ErrorCode.WRONG_URL_ERROR.value)
            except httpx.UnsupportedProtocol:
                error_text = f"无效的 url 协议（{url}）～请检查一下链接协议是否正确"
                if not args.batch:
                    error_text += (
                        "，如使用裸 id 功能，请确认该类型 id 是否支持当前单话模式，如不支持需要添加 `-b` 以使用批量模式"
                    )
                Logger.error(error_text)
                sys.exit(ErrorCode.WRONG_URL_ERROR.value)

            # 提取信息，构造解析任务～
            for extractor in extractors:
                if extractor.match(url):
                    download_list = await extractor(client, args)
                    break
            else:
                if args.batch:
                    # TODO: 指向文档中受支持的列表部分
                    Logger.error("url 不正确呦～")
                else:
                    Logger.error("url 不正确，也许该 url 仅支持批量下载，如果是这样，请使用参数 -b～")
                sys.exit(ErrorCode.WRONG_URL_ERROR.value)

            current_download_state = DownloadState.SKIP

            # 下载～
            for i, episode_data_coro in enumerate(download_list):
                if episode_data_coro is None:
                    continue

                # 中途校验，因为批量下载时可能会失效
                if not await validate_user_info({"is_login": args.login_strict, "vip_status": args.vip_strict}):
                    Logger.error("启用了严格校验大会员或登录模式，请检查 SESSDATA 或大会员状态！")
                    sys.exit(ErrorCode.NOT_LOGIN_ERROR.value)

                if current_download_state != DownloadState.SKIP and args.download_interval > 0:
                    Logger.info(f"下载间隔 {args.download_interval} 秒")
                    await sleep_with_status_bar_refresh(args.download_interval)

                # 这时候才真正开始解析链接
                episode_data = await episode_data_coro
                if episode_data is None:
                    continue
                # 保证路径唯一
                episode_data = ensure_unique_path(episode_data, unique_path)
                if args.batch:
                    Logger.custom(
                        f"{episode_data['filename']}",
                        Badge(f"[{i+1}/{len(download_list)}]", fore="black", back="cyan"),
                    )

                current_download_state = await start_downloader(
                    client,
                    episode_data,
                    {
                        "require_video": args.require_video,
                        "require_chapter_info": args.require_chapter_info,
                        "video_quality": args.video_quality,
                        "video_download_codec": args.vcodec.split(":")[0],
                        "video_save_codec": args.vcodec.split(":")[1],
                        "video_download_codec_priority": (
                            args.download_vcodec_priority.split(",")
                            if args.download_vcodec_priority != "auto"
                            else None
                        ),
                        "require_audio": args.require_audio,
                        "audio_quality": args.audio_quality,
                        "audio_download_codec": args.acodec.split(":")[0],
                        "audio_save_codec": args.acodec.split(":")[1],
                        "output_format": args.output_format,
                        "output_format_audio_only": args.output_format_audio_only,
                        "overwrite": args.overwrite,
                        "block_size": int(args.block_size * 1024 * 1024),
                        "num_workers": args.num_workers,
                        "metadata_format": {
                            "premiered": args.metadata_format_premiered,
                            "dateadded": TIME_FULL_FMT,
                        },
                        "banned_mirrors_pattern": args.banned_mirrors_pattern,
                    },
                )
                Logger.new_line()
            Logger.new_line()


def flatten_args(args: argparse.Namespace, parser: argparse.ArgumentParser) -> list[argparse.Namespace]:
    """递归展平列表参数"""
    args = copy.copy(args)
    validate_basic_arguments(args)
    # 查看是否存在于 alias 中
    alias_map = alias_parser(args.alias_file)
    if args.url in alias_map:
        args.url = alias_map[args.url]

    # 是否为下载列表
    if re.match(r"file://", args.url) or os.path.isfile(args.url):  # noqa: PTH113
        args_list: list[argparse.Namespace] = []
        # TODO: 如果是相对路径，需要相对于当前 list 路径
        for line in file_scheme_parser(args.url):
            local_args = parser.parse_args(line.split(), args)
            if local_args.no_inherit:
                local_args = parser.parse_args(line.split())
            Logger.debug(f"列表参数: {local_args}")
            args_list += flatten_args(local_args, parser)
        return args_list
    else:
        return [args]


def create_select_required_action(
    select: list[DownloadResourceType] | None = None, deselect: list[DownloadResourceType] | None = None
):
    selected_items = select or []
    deselected_items = deselect or []

    class SelectRequiredAction(argparse.Action):
        def __init__(self, option_strings: str, dest: str, nargs: int | str | None = None, **kwargs: Any):
            if nargs is not None:
                raise ValueError("nargs not allowed")
            super().__init__(option_strings, dest, nargs=0, **kwargs)

        def __call__(
            self,
            parser: argparse.ArgumentParser,
            namespace: argparse.Namespace,
            values: str | Sequence[str] | None,
            option_string: str | None = None,
        ):
            for select_item in selected_items:
                setattr(namespace, f"require_{select_item}", True)
            for deselect_item in deselected_items:
                setattr(namespace, f"require_{deselect_item}", False)

    return SelectRequiredAction


def invert_selection(select: list[DownloadResourceType]) -> list[DownloadResourceType]:
    return [tp for tp in DOWNLOAD_RESOURCE_TYPES if tp not in select]


def ensure_unique_path(episode_data: EpisodeData, unique_name_resolver: Callable[[str], str]) -> EpisodeData:
    original_filename = episode_data["filename"]
    new_name = unique_name_resolver(original_filename)
    episode_data["filename"] = new_name
    if original_filename != new_name:
        Logger.warning(f"文件名重复，已重命名为 {new_name}")
    return episode_data


if __name__ == "__main__":
    main()



from __future__ import annotations

import pytest

from yutto._typing import AId, AudioUrlMeta, CId, EpisodeId, SeasonId, VideoUrlMeta
from yutto.api.cheese import (
    get_cheese_list,
    get_cheese_playurl,
    get_season_id_by_episode_id,
)
from yutto.utils.fetcher import create_client
from yutto.utils.funcutils import as_sync


@pytest.mark.api
@as_sync
@pytest.mark.parametrize("episode_id", [EpisodeId("6945"), EpisodeId("6902")])
async def test_get_season_id_by_episode_id(episode_id: EpisodeId):
    season_id_excepted = SeasonId("298")
    async with create_client() as client:
        season_id = await get_season_id_by_episode_id(client, episode_id)
        assert season_id == season_id_excepted


@pytest.mark.api
@as_sync
async def test_get_cheese_title():
    season_id = SeasonId("298")
    async with create_client() as client:
        cheese_list = await get_cheese_list(client, season_id)
        title = cheese_list["title"]
        assert title == "林超：给年轻人的跨学科通识课"


@pytest.mark.api
@as_sync
async def test_get_cheese_list():
    season_id = SeasonId("298")
    async with create_client() as client:
        cheese_list = (await get_cheese_list(client, season_id))["pages"]
        assert cheese_list[0]["id"] == 1
        assert cheese_list[0]["name"] == "【先导片】给年轻人的跨学科通识课"
        assert cheese_list[0]["cid"] == CId("344779477")

        assert cheese_list[25]["id"] == 26
        assert cheese_list[25]["name"] == "回到真实世界（下）"
        assert cheese_list[25]["cid"] == CId("506369050")


@pytest.mark.api
@pytest.mark.ci_skip
@as_sync
async def test_get_cheese_playurl():
    avid = AId("545852212")
    episode_id = EpisodeId("6902")
    cid = CId("344779477")
    async with create_client() as client:
        playlist: tuple[list[VideoUrlMeta], list[AudioUrlMeta]] = await get_cheese_playurl(
            client, avid, episode_id, cid
        )
        assert len(playlist[0]) > 0
        assert len(playlist[1]) > 0


@pytest.mark.api
@as_sync
async def test_get_cheese_subtitles():
    # TODO: 暂未找到需要字幕的课程
    pass



from typing import ClassVar

class DanmakuElem:
    @property
    def id(self) -> int: ...
    @property
    def progress(self) -> int: ...
    @property
    def mode(self) -> int: ...
    @property
    def fontsize(self) -> int: ...
    @property
    def color(self) -> int: ...
    @property
    def mid_hash(self) -> str: ...
    @property
    def content(self) -> str: ...
    @property
    def ctime(self) -> int: ...
    @property
    def action(self) -> str: ...
    @property
    def pool(self) -> int: ...
    @property
    def id_str(self) -> str: ...
    @property
    def attr(self) -> int: ...
    @property
    def animation(self) -> str: ...

class DmSegMobileReply:
    @property
    def elems(self) -> list[DanmakuElem]: ...
    @staticmethod
    def decode(data: bytes) -> DmSegMobileReply: ...

class CommentPosition:
    Scroll: ClassVar[CommentPosition]
    Top: ClassVar[CommentPosition]
    Bottom: ClassVar[CommentPosition]
    Reversed: ClassVar[CommentPosition]
    Special: ClassVar[CommentPosition]

    @property
    def id(self) -> int: ...

class Comment:
    timeline: float
    timestamp: int
    no: int
    comment: str
    pos: CommentPosition
    color: int
    size: float
    height: float
    width: float

def read_comments_from_xml(text: str, fontsize: float) -> list[Comment]: ...
def read_comments_from_protobuf(data: bytes, fontsize: float) -> list[Comment]: ...
def parse_special_comment(
    content: str, zoom_factor: tuple[float, float, float]
) -> tuple[tuple[int, int, float, float, float, float], int, int, str, int, float, int, str, bool]: ...
def xml_to_ass(
    inputs: list[str],
    stage_width: int,
    stage_height: int,
    reserve_blank: int,
    font_face: str,
    font_size: float,
    text_opacity: float,
    duration_marquee: float,
    duration_still: float,
    comment_filter: list[str],
    is_reduce_comments: bool,
) -> str: ...
def protobuf_to_ass(
    inputs: list[bytes],
    stage_width: int,
    stage_height: int,
    reserve_blank: int,
    font_face: str,
    font_size: float,
    text_opacity: float,
    duration_marquee: float,
    duration_still: float,
    comment_filter: list[str],
    is_reduce_comments: bool,
) -> str: ...
def get_danmaku_meta_size(buffer: bytes) -> int: ...



from __future__ import annotations

from typing import TypeVar, cast

from biliass._core import (
    Comment,
    protobuf_to_ass,
    read_comments_from_protobuf,
    read_comments_from_xml,
    xml_to_ass,
)

T = TypeVar("T")


def read_comments_bilibili_xml(text: str | bytes, fontsize: float) -> list[Comment]:
    if isinstance(text, bytes):
        text = text.decode()
    return read_comments_from_xml(text, fontsize)


def read_comments_bilibili_protobuf(protobuf: bytes | str, fontsize: float) -> list[Comment]:
    assert isinstance(protobuf, bytes), "protobuf supports bytes only"
    return read_comments_from_protobuf(protobuf, fontsize)


def Danmaku2ASS(
    inputs: list[str | bytes] | str | bytes,
    stage_width: int,
    stage_height: int,
    input_format: str = "xml",
    reserve_blank: int = 0,
    font_face: str = "sans-serif",
    font_size: float = 25.0,
    text_opacity: float = 1.0,
    duration_marquee: float = 5.0,
    duration_still: float = 5.0,
    comment_filter: str | None = None,
    is_reduce_comments: bool = False,
) -> str:
    comment_filters: list[str] = [comment_filter] if comment_filter is not None else []
    if not isinstance(inputs, list):
        inputs = [inputs]

    if input_format == "xml":
        inputs = [text if isinstance(text, str) else text.decode() for text in inputs]
        return xml_to_ass(
            cast(list[str], inputs),
            stage_width,
            stage_height,
            reserve_blank,
            font_face,
            font_size,
            text_opacity,
            duration_marquee,
            duration_still,
            comment_filters,
            is_reduce_comments,
        )
    elif input_format == "protobuf":
        for input in inputs:
            if isinstance(input, str):
                raise ValueError("Protobuf can only be read from bytes")
        return protobuf_to_ass(
            cast(list[bytes], inputs),
            stage_width,
            stage_height,
            reserve_blank,
            font_face,
            font_size,
            text_opacity,
            duration_marquee,
            duration_still,
            comment_filters,
            is_reduce_comments,
        )
    else:
        raise TypeError(f"Invalid input format {input_format}")



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from biliass import get_danmaku_meta_size

from yutto.api.user_info import get_user_info
from yutto.utils.fetcher import Fetcher

if TYPE_CHECKING:
    import httpx

    from yutto._typing import AvId, CId
    from yutto.utils.danmaku import DanmakuData, DanmakuSaveType


async def get_xml_danmaku(client: httpx.AsyncClient, cid: CId) -> str:
    danmaku_api = "http://comment.bilibili.com/{cid}.xml"
    results = await Fetcher.fetch_text(client, danmaku_api.format(cid=cid), encoding="utf-8")
    assert results is not None
    return results


async def get_protobuf_danmaku_segment(client: httpx.AsyncClient, cid: CId, segment_id: int = 1) -> bytes:
    danmaku_api = "http://api.bilibili.com/x/v2/dm/web/seg.so?type=1&oid={cid}&segment_index={segment_id}"
    results = await Fetcher.fetch_bin(client, danmaku_api.format(cid=cid, segment_id=segment_id))
    assert results is not None
    return results


async def get_protobuf_danmaku(client: httpx.AsyncClient, avid: AvId, cid: CId) -> list[bytes]:
    danmaku_meta_api = "https://api.bilibili.com/x/v2/dm/web/view?type=1&oid={cid}&pid={aid}"
    aid = avid.as_aid()
    meta_results = await Fetcher.fetch_bin(client, danmaku_meta_api.format(cid=cid, aid=aid.value))
    assert meta_results is not None
    size = get_danmaku_meta_size(meta_results)

    results = await asyncio.gather(
        *[get_protobuf_danmaku_segment(client, cid, segment_id) for segment_id in range(1, size + 1)]
    )
    return results


async def get_danmaku(
    client: httpx.AsyncClient,
    cid: CId,
    avid: AvId,
    save_type: DanmakuSaveType,
) -> DanmakuData:
    # 在已经登录的情况下，使用 protobuf，因为未登录时 protobuf 弹幕会少非常多
    source_type = "xml" if save_type == "xml" or not (await get_user_info(client))["is_login"] else "protobuf"
    danmaku_data: DanmakuData = {
        "source_type": source_type,
        "save_type": save_type,
        "data": [],
    }

    if source_type == "xml":
        danmaku_data["data"].append(await get_xml_danmaku(client, cid))
    else:
        danmaku_data["data"].extend(await get_protobuf_danmaku(client, avid, cid))
    return danmaku_data



from __future__ import annotations

import pytest

from yutto._typing import BvId, CId, EpisodeId, MediaId, SeasonId
from yutto.api.bangumi import (
    get_bangumi_list,
    get_bangumi_playurl,
    get_bangumi_subtitles,  # type: ignore
    get_season_id_by_episode_id,
    get_season_id_by_media_id,
)
from yutto.utils.fetcher import create_client
from yutto.utils.funcutils import as_sync


@pytest.mark.api
@as_sync
async def test_get_season_id_by_media_id():
    media_id = MediaId("28223066")
    season_id_excepted = SeasonId("28770")
    async with create_client() as client:
        season_id = await get_season_id_by_media_id(client, media_id)
        assert season_id == season_id_excepted


@pytest.mark.api
@as_sync
@pytest.mark.parametrize("episode_id", [EpisodeId("314477"), EpisodeId("300998")])
async def test_get_season_id_by_episode_id(episode_id: EpisodeId):
    season_id_excepted = SeasonId("28770")
    async with create_client() as client:
        season_id = await get_season_id_by_episode_id(client, episode_id)
        assert season_id == season_id_excepted


@pytest.mark.api
@as_sync
async def test_get_bangumi_title():
    season_id = SeasonId("28770")
    async with create_client() as client:
        title = (await get_bangumi_list(client, season_id))["title"]
        assert title == "我的三体之章北海传"


@pytest.mark.api
@as_sync
async def test_get_bangumi_list():
    season_id = SeasonId("28770")
    async with create_client() as client:
        bangumi_list = (await get_bangumi_list(client, season_id))["pages"]
        assert bangumi_list[0]["id"] == 1
        assert bangumi_list[0]["name"] == "第1话"
        assert bangumi_list[0]["cid"] == CId("144541892")
        assert bangumi_list[0]["metadata"] is not None
        assert bangumi_list[0]["metadata"]["title"] == "第1话"

        assert bangumi_list[8]["id"] == 9
        assert bangumi_list[8]["name"] == "第9话"
        assert bangumi_list[8]["cid"] == CId("162395026")
        assert bangumi_list[8]["metadata"] is not None
        assert bangumi_list[8]["metadata"]["title"] == "第9话"


@pytest.mark.api
@pytest.mark.ci_skip
@as_sync
async def test_get_bangumi_playurl():
    avid = BvId("BV1q7411v7Vd")
    cid = CId("144541892")
    async with create_client() as client:
        playlist = await get_bangumi_playurl(client, avid, cid)
        assert len(playlist[0]) > 0
        assert len(playlist[1]) > 0


@pytest.mark.api
@as_sync
async def test_get_bangumi_subtitles():
    # TODO: 暂未找到需要字幕的番剧（非港澳台）
    pass



from __future__ import annotations

import asyncio

import httpx
import pytest

from yutto.processor.downloader import slice_blocks
from yutto.utils.asynclib import CoroutineWrapper
from yutto.utils.fetcher import Fetcher, create_client
from yutto.utils.file_buffer import AsyncFileBuffer
from yutto.utils.funcutils import as_sync

from ..conftest import TEST_DIR


@pytest.mark.processor
@as_sync
async def test_150_kB_downloader():
    # test_dir = "./downloader_test/"
    # url = "https://file-examples-com.github.io/uploads/2017/04/file_example_MP4_480_1_5MG.mp4"
    # 因为 file-examples-com 挂掉了（GitHub 账号都消失了，因此暂时使用一个别处的 mirror）
    url = "https://github.com/nhegde610/samples-files/raw/main/file_example_MP4_480_1_5MG.mp4"
    file_path = TEST_DIR / "test_150_kB.pdf"
    async with await AsyncFileBuffer(file_path, overwrite=False) as buffer:
        async with create_client(
            timeout=httpx.Timeout(7, connect=3),
        ) as client:
            Fetcher.set_semaphore(4)
            size = await Fetcher.get_size(client, url)
            coroutines = [
                CoroutineWrapper(Fetcher.download_file_with_offset(client, url, [], buffer, offset, block_size))
                for offset, block_size in slice_blocks(buffer.written_size, size, 1 * 1024 * 1024)
            ]

            print("开始下载……")
            await asyncio.gather(*coroutines)
            print("下载完成！")
            assert size == file_path.stat().st_size, "文件大小与实际大小不符"


@pytest.mark.processor
@as_sync
async def test_150_kB_no_slice_downloader():
    # test_dir = "./downloader_test/"
    # url = "https://file-examples-com.github.io/uploads/2017/04/file_example_MP4_480_1_5MG.mp4"
    url = "https://github.com/nhegde610/samples-files/raw/main/file_example_MP4_480_1_5MG.mp4"
    file_path = TEST_DIR / "test_150_kB_no_slice.pdf"
    async with await AsyncFileBuffer(file_path, overwrite=False) as buffer:
        async with create_client(
            timeout=httpx.Timeout(7, connect=3),
        ) as client:
            Fetcher.set_semaphore(4)
            size = await Fetcher.get_size(client, url)
            coroutines = [CoroutineWrapper(Fetcher.download_file_with_offset(client, url, [], buffer, 0, size))]

            print("开始下载……")
            await asyncio.gather(*coroutines)
            print("下载完成！")
            assert size == file_path.stat().st_size, "文件大小与实际大小不符"
